#!/usr/bin/env python3
# coding: utf-8

"""
Mediaview, "Diary of a madman" (https://git.io/fATFM, https://git.io/fATNI).
Usage:

    $ python mediaview.py [FILE]

Fields:
    line_channel,    // 0
    line_topic,      // 1
    title,           // 2
    ,                // 3
    ,                // 4
    hr_duration,     // 5
    size,            // 6
    description,     // 7
    url_video,       // 8
    url_website,     // 9
    url_subtitle,    // 10
    ,                // 11
    url_video_low,   // 12
    ,                // 13
    url_video_hd,    // 14
    ,                // 15
    timestamp        // 16

The url_video_low, url_video_hd are encoded like "offset|s", where the final
url will be: url_video[offset] + s

Updates (diff) contain a few more fields.
"""

from __future__ import print_function

import collections
import re
import sys
import json
import marcx
import base64
import hashlib
import datetime
import tqdm

channels = ("3Sat", "ARD", "ARTE.DE", "ARTE.FR", "BR", "DW", "HR", "KiKA",
            "MDR", "NDR", "ORF", "PHOENIX", "RBB", "SR", "SRF", "SRF.Podcast",
            "SWR", "WDR", "ZDF", "ZDF-tivi", "3sat")

records = collections.defaultdict(dict)
seen = set()

inputfilename = "169_input.json"
outputfilename = "169_output.mrc"

if len(sys.argv) == 3:
    inputfilename, outputfilename = sys.argv[1:]

inputfile = open(inputfilename, "r")
outputfile = open(outputfilename, "wb")
content = inputfile.read()

pattern = re.compile(r"""^{"Filmliste":|,"X":|}$""")
lines = pattern.split(content)

for line in lines:

    # sort -u dups.ndj > uniq.ndj
    if line in seen:
        print('dropping duplicate line: %s ...' % line[:40], file=sys.stderr)
        continue
    # XXX: up to a few 100000 lines this is ok
    seen.add(line)

    try:
        doc = json.loads(line)
    except Exception as exc:
        print(exc, file=sys.stderr)
    else:
        if doc[0] != "":
            current_channel = doc[0]
        if doc[1] != "":
            current_topic = doc[1]
        record = {
            "channel": current_channel,
            "topic": current_topic,
            "title": doc[2],
            "hr_duration": doc[5],
            "size": doc[6],
            "description": doc[7],
            "url_video": doc[8],
            "url_website": doc[9],
            "url_subtitle": doc[10],
            "url_video_low": doc[12],
            "url_video_hd": doc[14],
            "timestamp": doc[16],
        }

        try:
            if record["url_video_low"] and "|" in record["url_video_low"]:
                offset, suffix = record["url_video_low"].split("|")
                record['url_video_low'] = record['url_video'][:int(offset)] + suffix
        except Exception as exc:
            print(exc, file=sys.stderr)

        try:
            if record["url_video_hd"] and "|" in record["url_video_hd"]:
                offset, suffix = record["url_video_hd"].split("|")
                record['url_video_hd'] = record['url_video'][:int(offset)] + suffix
        except Exception as exc:
            print(exc, file=sys.stderr)

        if record["timestamp"] == "":
            continue

        hash = hashlib.sha1()
        fields = record["channel"] + record["topic"] + record["title"] + record["timestamp"]
        hash.update(fields.encode("utf-8").strip())
        hash_record = hash.hexdigest()

        f001 = record["channel"] + record["topic"][:10] + record["title"][:20] + record["title"][-20:] + record["hr_duration"] + record["size"] + record["timestamp"]
        f001 = f001.encode("utf-8")
        f001 = base64.urlsafe_b64encode(f001)
        f001 = f001.decode("utf-8").rstrip("=")
        records[hash_record].update({"f001": f001})

        f245a = record["title"]
        records[hash_record].update({"f245a": f245a})

        timestamp = record["timestamp"]
        if timestamp:
            timestamp = int(timestamp)
            f260c = datetime.datetime.fromtimestamp(timestamp).strftime("%Y")
        else:
            f260c = ""
        publisher = ["b", record["channel"] + ", ", "c", f260c]
        records[hash_record].update({"f260": publisher})

        f306a = record["hr_duration"]
        records[hash_record].update({"f306a": f306a})

        if record["topic"] not in channels and " / " not in record["topic"] and record["topic"] != record["title"]:
            records[hash_record].update({"f490a": record["topic"]})

        timestamp = record["timestamp"]
        if timestamp:
            timestamp = int(timestamp)
            f500a = datetime.datetime.fromtimestamp(timestamp).strftime("Gesendet am %d.%m.%Y um %H:%M Uhr")
            if "00:00" in f500a:
                f500a = f500a.replace(" um 00:00 Uhr", "")
            records[hash_record].update({"f500a": f500a})

        f520a = record["description"]
        desc = records[hash_record].setdefault("f520a", [])
        if f520a not in desc:
            records[hash_record].setdefault("f520a", []).append(f520a)

        if record["url_website"]:
            sites = records[hash_record].setdefault("website", [])
            if record["url_website"] not in sites:
                records[hash_record].setdefault("website", []).append(record["url_website"])
        if record["url_video_low"]:
            records[hash_record].setdefault("low", []).append(record["url_video_low"])
        if record["url_video"]:
            records[hash_record].setdefault("medium", []).append(record["url_video"])
        if record["url_video_hd"]:
            records[hash_record].setdefault("high", []).append(record["url_video_hd"])

for record in tqdm.tqdm(records.values(), total=len(records)):

    marcrecord = marcx.Record(force_utf8=True)
    marcrecord.strict = False
    marcrecord.leader = "     cam  22        4500"
    marcrecord.add("001", data="finc-169-" + record["f001"])
    marcrecord.add("007", data="cr")
    marcrecord.add("008", data="050228s%s r 000 0 ger d" % record["f260"][3])
    marcrecord.add("245", a=record["f245a"])
    marcrecord.add("260", subfields=record["f260"])
    marcrecord.add("306", a=record["f306a"])
    marcrecord.add("490", a=record.get("f490a"))
    marcrecord.add("500", a=record["f500a"])

    for i, f520a in enumerate(record.get("f520a", []), start=1):
        if len(record["f520a"]) == 1:
            label = ""
        else:
            label = "Video " + str(i) + ":\n"
        marcrecord.add("520", a=label+f520a)

    for i, url in enumerate(record.get("website", []), start=1):
        if len(record["website"]) == 1:
            i = ""
        marcrecord.add("856", q="text/html", _3="Link zur Webseite %s" % i, u=url)

    for i, url in enumerate(record.get("low", []), start=1):
        if len(record["low"]) == 1:
            i = ""
        marcrecord.add("856", q="text/html", _3="Link zu Video %s (LD)" % i, u=url)

    for i, url in enumerate(record.get("medium", []), start=1):
        if len(record["medium"]) == 1:
            i = ""
        marcrecord.add("856", q="text/html", _3="Link zu Video %s (SD)" % i, u=url)

    for i, url in enumerate(record.get("high", []), start=1):
        if len(record["high"]) == 1:
            i = ""
        marcrecord.add("856", q="text/html", _3="Link zu Video %s (HD)" % i, u=url)

    marcrecord.add("935", b="cofz", c="vide")
    subfields = ["a", f001, "b", "169", "c", "sid-169-col-mediathek"]
    marcrecord.add("980", subfields=subfields)

    outputfile.write(marcrecord.as_marc())

inputfile.close()
outputfile.close()



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Lynda, refs #11477.

* https://www.lynda.com/courselist?library=en&retired=true
* https://www.lynda.com/courselist?library=de&retired=true

"""

import csv
import datetime
import json

import luigi

from gluish.format import TSV, Gzip
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.common import FTPMirror
from siskin.decorator import deprecated
from siskin.task import DefaultTask

class LyndaTask(DefaultTask):
    """
    Base class for lynda.com courses.
    """
    TAG = '141'

    def closest(self):
        return monthly(date=self.date)

class LyndaPaths(LyndaTask):
    """
    Mirror SLUB FTP.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    max_retries = luigi.IntParameter(default=10, significant=False)
    timeout = luigi.IntParameter(default=20, significant=False,
                                 description='timeout in seconds')

    def requires(self):
        return FTPMirror(host=self.config.get('lynda', 'ftp-host'),
                         base=self.config.get('lynda', 'ftp-base'),
                         username=self.config.get('lynda', 'ftp-username'),
                         password=self.config.get('lynda', 'ftp-password'),
                         pattern=self.config.get('lynda', 'ftp-pattern'),
                         max_retries=self.max_retries,
                         timeout=self.timeout)

    def run(self):
        self.input().move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class LyndaIntermediateSchema(LyndaTask):
    """
    XXX: Workaround SOLR, refs #11477.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return LyndaPaths(date=self.date)

    def run(self):
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                if row.path.endswith("latest"):
                    output = shellout(""" gunzip -c {input} |
                                      jq -rc '.fullrecord' |
                                      jq -rc 'del(.["x.labels"])' |
                                      jq -rc '. + {{"finc.id": .["finc.record_id"]}}' | gzip -c > {output} """,
                                      input=row.path)
                    luigi.LocalTarget(output).move(self.output().path)
                    break
            else:
                raise RuntimeError("no latest symlink found in folder")

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ldj.gz"), format=Gzip)

class LyndaDownloadDeprecated(LyndaTask):
    """
    Download list.

    Deprecated: Use SLUB processed version via FTP download.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    language = luigi.Parameter(default="en",
                               description="available languages: de, en")

    @deprecated
    def run(self):
        url = "https://www.lynda.com/courselist?library=%s&retired=true" % self.language
        output = shellout("""curl --fail "{url}" > {output}""", url=url)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='zip'))

class LyndaIntermediateSchemaDeprecated(LyndaTask):
    """
    Intermediate schema conversion.

    Deprecated: Use SLUB processed version via FTP download.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    language = luigi.Parameter(default="en",
                               description="available languages: de, en")

    def requires(self):
        return LyndaDownloadDeprecated(language=self.language, date=self.date)

    @deprecated
    def run(self):
        """
        It's a Zip with a single file.
        """
        output = shellout("""unzip -p {input} | dos2unix > {output}""",
                          input=self.input().path)

        language_map = {'en': 'eng', 'de': 'deu'}
        collection_map = {'en': 'English', 'de': 'Deutsch'}

        with open(output) as csvfile:
            reader = csv.DictReader(csvfile)
            with self.output().open('w') as output:
                for row in reader:
                    # ['Course ID', 'Course Title', 'Author', 'Release Date',
                    # 'Level', 'Duration', 'Category', 'Categories',
                    # 'Software', 'Description', 'Course URL', 'AICC URL', 'SSO
                    # URL', 'Small Thumbnail', 'Medium Thumbnail', 'Large
                    # Thumbnail', 'Status', 'Retire Date', 'Library']

                    # 02/13/2015
                    date = datetime.datetime.strptime(
                        row['Release Date'], '%m/%d/%Y')

                    categories = set([row['Category'], row['Categories']])
                    parts = [s.strip() for s in row['Course Title'].split(':')]

                    if len(parts) == 1:
                        title, subtitle = parts[0], ''
                    if len(parts) > 1:
                        title, subtitle = parts[0], ': '.join(parts[1:])

                    doc = {
                        'finc.id': 'ai-141-%s' % row['Course ID'],
                        'finc.record_id': row['Course ID'],
                        'finc.source_id': '141',
                        'finc.format': 'ElectronicResourceRemoteAccess',
                        'finc.mega_collection': 'Lynda.com %s' % (collection_map[self.language]),
                        'rft.genre': 'document',
                        'rft.pub': 'Lynda',
                        'rft.date': date.strftime("%Y-%m-%d"),
                        'x.date': date.strftime('%Y-%m-%dT%H:%M:%SZ'),
                        'ris.type': 'VIDEO',
                        'languages': [language_map[self.language]],
                        'x.subjects': list(categories),
                        'rft.atitle': title,
                        'rft.subtitle': subtitle,
                        'rft.abstract': row['Description'],
                        'rft.url': [row['Course URL']],
                        'rft.author': [{'rft.au': row['Author']}],
                    }
                    output.write(json.dumps(doc))
                    output.write("\n")

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))



# coding: utf-8
# pylint: disable=C0301,E1101

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Datenbank Internetquellen, #5374, #9913, #11005.

Fixed source from the defunct http://www.medien-buehne-film.de/.

About 214M in 163 XML files, no schema, root XML element is called DBClear,
currently the files are named monday.xml.part133, monday.xml.part134, and so
on.

Configuration
-------------

[dbinet]

directory = /path/to/dir/containing/raw/files

"""

import datetime
import glob
import json
import os
import tempfile

import luigi
import pytz
import requests
from luigi.format import Gzip

from gluish.format import TSV
from gluish.utils import shellout
from siskin.task import DefaultTask
from siskin.utils import load_set_from_file


class DBInetTask(DefaultTask):
    """
    Base task, refs #5374, #9913, #11005, #13653.
    """
    TAG = "80"


class DBInetFiles(DBInetTask):
    """
    A list of raw files.
    """

    def run(self):
        directory = self.config.get('dbinet', 'directory')
        with self.output().open('w') as output:
            for path in sorted(glob.glob(os.path.join(directory, "*xml.part*"))):
                output.write_tsv(path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist'), format=TSV)


class DBInetJSON(DBInetTask):
    """
    Combine, prepare, requires xlstproc, sed, jq.
    """

    def requires(self):
        return DBInetFiles()

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                shellout("""python {xmltojson} <(xsltproc {xsl} <(sed -e 's/&#.;//g' -e 's/&#..;//g' {input})) /dev/stdout |
                            jq -cr ".list[][]" >> {output} """,
                         xmltojson=self.assets('80/xmltojson.py'),
                         xsl=self.assets('80/convert.xsl'), input=row.path, output=stopover)

        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))


class DBInetIntermediateSchema(DBInetTask):
    """
    Convert via jq.

    This tasks can create a list of failed documents, with `--failed errdocs.json` flag.

    To create file containing link check results for failed links, run:

        $ taskrm DBInetIntermediateSchema
        $ taskdo DBInetIntermediateSchema --failed errdocs.json
        $ jq -rc '.url[]' errdocs.json | clinker | jq -r 'select(.status != 200) | .link' > urls_to_drop.txt

    Put path to urls_to_drop.txt (one URL per line) into:

        [dbinet]

        exclude-links = urls_to_drop.txt

    """

    urlcheck = luigi.BoolParameter(default=False, significant=False)
    failed = luigi.Parameter(default="", description="write failed JSON to this file", significant=False)

    def requires(self):
        return DBInetJSON()

    def run(self):
        """
        When URLs are checked, only write records with at least one reachable URL.
        """
        output = shellout("jq -rc -f {filter} {input} | pigz -c > {output}",
                          filter=self.assets('80/filter.jq'), input=self.input().path)

        # If we have file to a list of links to exclude, then load this list.
        urls_to_drop = set()

        exclude_file = self.config.get("dbinet", "exclude-links")
        if exclude_file is None or not os.path.exists(exclude_file):
            self.logger.debug(self.__doc__)
        else:
            urls_to_drop = load_set_from_file(exclude_file)

        # Data cleaning.
        dropped_records = 0

        with luigi.LocalTarget(output, format=Gzip).open() as handle:
            with luigi.LocalTarget(is_tmp=True, format=Gzip).open('w') as tmp:
                for i, line in enumerate(handle):
                    line = line.strip()
                    if not line:
                        continue
                    doc = json.loads(line)

                    # Mitigate stutter in result list, refs #13653.
                    series = doc["rft.series"]
                    if series:
                        doc["x.footnotes"] = [doc["rft.series"]]
                        doc["rft.series"] = ""

                    if doc.get("rft.date") in (None, "", "null"):
                        # This is probably a web site.
                        if self.failed:
                            with open(self.failed, "a") as failed:
                                failed.write(json.dumps(doc))
                                failed.write("\n")

                        # If the link is valid, then add current date and
                        # change document format.
                        for url in doc.get("url", []):
                            if url in urls_to_drop:
                                self.logger.debug("dropping document, since link is in exclude list: %s", url)
                                dropped_records += 1
                                continue

                        today = datetime.datetime.now(pytz.timezone('UTC'))
                        doc["rft.date"] = today.strftime("%Y-%m-%d")
                        doc["x.date"] = today.isoformat()
                        doc["finc.format"] = "ElectronicIntegratingResource"
                        doc["finc.genre"] = "document"

                    doc["rft.place"] = [v for v in doc["rft.place"] if v]
                    doc["rft.pub"] = [v for v in doc["rft.pub"] if v]

                    tmp.write(json.dumps(doc))
                    tmp.write("\n")

        luigi.LocalTarget(tmp.path).move(self.output().path)
        self.logger.debug("dropped %s records during cleaning", dropped_records)

        if self.urlcheck:
            raise NotImplementedError()

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)



# coding: utf-8

"""
A few tests for the mab module, refs #8392.
"""

import os
import tempfile

import pytest

from siskin.mab import MabXMLFile

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

# Example MAB file with one record.
sample_file_one = """<?xml version="1.0" encoding="UTF-8"?>
<datei xmlns="http://www.ddb.de/professionell/mabxml/mabxml-1.xsd" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.ddb.de/professionell/mabxml/mabxml-1.xsd http://www.d-nb.de/standardisierung/formate/mabxml-1.xsd">
<datensatz typ="u" status="n" mabVersion="M2.0" xmlns="http://www.ddb.de/professionell/mabxml/mabxml-1.xsd"><feld nr="001" ind=" ">10598985</feld><feld nr="002" ind="a">20070926</feld><feld nr="003" ind=" ">20160311</feld><feld nr="004" ind=" ">20190110</feld><feld nr="010" ind=" ">10390691</feld><feld nr="020" ind="b">ITKGDK500164319</feld><feld nr="030" ind=" ">e|5|xr|zc||||</feld><feld nr="050" ind=" ">a|a|||||||||||</feld><feld nr="051" ind=" ">s|||||||</feld><feld nr="070" ind="a">GDK</feld><feld nr="071" ind=" ">80</feld><feld nr="081" ind=" ">81875804</feld><feld nr="082" ind=" ">I. 9 I 61</feld><feld nr="083" ind=" ">Verfügbar</feld><feld nr="084" ind=" ">Magazin</feld><feld nr="089" ind=" ">2006</feld><feld nr="090" ind=" ">42006</feld><feld nr="335" ind=" ">FIAF 2007 Tokyo</feld><feld nr="425" ind=" ">2007</feld><feld nr="425" ind="a">2007</feld><feld nr="433" ind=" ">[ca. 300  Seiten ]</feld><feld nr="501" ind=" ">Enthält die Berichte über Aktivitäten im Jahr 2006</feld></datensatz>
</datei>
"""

# Example MAB file with two records.
sample_file_two = """<?xml version="1.0" encoding="UTF-8"?>
<datei xmlns="http://www.ddb.de/professionell/mabxml/mabxml-1.xsd" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.ddb.de/professionell/mabxml/mabxml-1.xsd http://www.d-nb.de/standardisierung/formate/mabxml-1.xsd">
  <datensatz xmlns="http://www.ddb.de/professionell/mabxml/mabxml-1.xsd" typ="h" status="n" mabVersion="M2.0">
    <feld nr="001" ind=" ">000062947</feld>
    <feld nr="002" ind="a">20170704</feld>
    <feld nr="003" ind=" ">20170906</feld>
    <feld nr="030" ind=" ">z|zuar|z||||7</feld>
    <feld nr="036" ind="a">CN</feld>
    <feld nr="037" ind="b">eng</feld>
    <feld nr="050" ind=" ">a|a|||||||||||</feld>
    <feld nr="060" ind=" ">
      <uf code="b">txt</uf>
    </feld>
    <feld nr="100" ind=" ">Zhang, Yisheng</feld>
    <feld nr="102" ind=" ">000021597</feld>
    <feld nr="104" ind="b">Ma, Mingtu</feld>
    <feld nr="106" ind=" ">000021596</feld>
    <feld nr="331" ind=" ">Advanced High Strength Steel and Press Hardening [ICHSU2015]</feld>
    <feld nr="335" ind=" ">proceedings of the2nd International Conference on Advanced High Strength Steel and Press Hardening (ICHSU2015)</feld>
    <feld nr="359" ind=" ">Editors Yisheng Zhang (Huazhong University of Science and Technology, China), Mingtu Ma (China Automotive Engineering Research Institute, China).</feld>
    <feld nr="419" ind=" ">
      <uf code="a">Singapur</uf>
      <uf code="b">World Scientific Publishing</uf>
      <uf code="c">2016</uf>
    </feld>
    <feld nr="433" ind=" ">XV, 736 Seiten</feld>
    <feld nr="434" ind=" ">Illustrationen, Tabellen</feld>
    <feld nr="435" ind=" ">24 cm</feld>
    <feld nr="540" ind="a">ISBN 978-981-3140-61-5</feld>
    <feld nr="655" ind="e">
      <uf code="u">http://www2.stahl-online.de/Bibl-Dokumente.nsf/af8dcd050f6931fdc1257798003b85a6/227fdc3019e0cf3cc12581530032b592/$FILE/4.6745.pdf</uf>
    </feld>
    <feld nr="675" ind=" ">ICHSU 2015</feld>
    <feld nr="710" ind=" ">hochfester Stahl</feld>
    <feld nr="710" ind=" ">Blechumformen</feld>
    <feld nr="710" ind=" ">Pressh&#xE4;rten</feld>
  </datensatz>
  <datensatz>
    <!-- record #2 -->
  </datensatz>
</datei>
"""

def test_from_filelike():
    """
    Test, whether we can load from a filelike object.
    """
    filelike = StringIO(sample_file_two)
    mabf = MabXMLFile(filelike)
    assert len([_ for _ in mabf]) == 2

def test_from_file():
    """
    Test, whether we can load from a filename.
    """
    with tempfile.NamedTemporaryFile(delete=False, mode='w') as f:
        f.write(sample_file_two)

    mabf = MabXMLFile(f.name)
    assert len([_ for _ in mabf]) == 2

    os.remove(f.name)


def test_invalid_input():
    """
    A few invalid inputs, there are many more.
    """
    inputs = (
        """<xml></xml>""",
        """<?xml version="1.0" encoding="UTF-8"?>
        <BadRoot></BadRoot>
        """,
        """<a></a>""",
    )

    for s in inputs:
        with pytest.raises(ValueError):
            MabXMLFile(s)

def test_valid_empty_file():
    """
    Ok, but empty.
    """
    inputs = (
        """<?xml version="1.0" encoding="UTF-8"?>
        <datei></datei>
        """,
        """<?xml version="1.0" encoding="UTF-8"?>
        <datei xmlns="http://www.ddb.de/professionell/mabxml/mabxml-1.xsd" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.ddb.de/professionell/mabxml/mabxml-1.xsd http://www.d-nb.de/standardisierung/formate/mabxml-1.xsd"></datei>
        """,
        """<?xml version="1.0" encoding="UTF-8"?>
        <datei xmlns="http://www.ddb.de/professionell/mabxml/mabxml-1.xsd" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.ddb.de/professionell/mabxml/mabxml-1.xsd http://www.d-nb.de/standardisierung/formate/mabxml-1.xsd">
        <a></a>
        </datei>
        """
    )
    for s in inputs:
        assert MabXMLFile(s) is not None

def test_record_iteration():
    """
    Whether we can iterate over the file object.
    """
    inputs = (
        (sample_file_one, 1),
        (sample_file_two, 2),
    )
    for s, count in inputs:
        mabf = MabXMLFile(s)
        assert len([_ for _ in mabf]) == count

def test_record_field():
    """
    Access to first value in fields.
    """
    mabf = MabXMLFile(sample_file_two)
    record = mabf.next()
    assert record.field("36") is None
    assert record.field("036") == "CN"
    assert record.field("060") is None
    assert record.field("710") == "hochfester Stahl"

def test_record_fields():
    """
    Access to multiple values.
    """
    mabf = MabXMLFile(sample_file_two)
    record = mabf.next()
    assert len(record.fields("710")) == 3
    assert len(record.fields("999")) == 0

def test_record_subfield():
    """
    Also known as Unterfeld.
    """
    mabf = MabXMLFile(sample_file_two)
    record = mabf.next()
    assert record.field("419") is None
    assert record.field("419", code="c") == "2016"
    assert record.field("419", "c") == "2016"



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Robert Schenk, <robert.schenk@uni-leipzig.de>
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
VKFilm Bundesarchiv, #8695.

TODO: Unify VKFilm* tasks.

Config
------

[vkfilmba]

baseurl = https://example.com

data = http://example.com/data.zip
deletions = http://example.com/del.txt

"""

import datetime
import os
import re
import shutil
import tempfile
import zipfile

import luigi

import pymarc
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.task import DefaultTask
from siskin.utils import iterfiles


class VKFilmBATask(DefaultTask):
    """
    Bundesarchiv.
    """
    TAG = '148'

    def closest(self):
        return weekly(date=self.date)


class VKFilmBADownload(VKFilmBATask):
    """
    Download raw data daily. At least try, files may be missing, refs #12460.
    """
    date = luigi.DateParameter(default=datetime.date.today())

    def run(self):
        """
        XXX: Maybe ensure zip.
        """
        output = shellout("""curl --fail "{url}" > {output} """,
                          url=self.config.get('vkfilmba', 'data'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        """
        We expect zip, but to guarantee.
        """
        return luigi.LocalTarget(path=self.path(ext="zip"))


class VKFilmBADownloadDeletions(VKFilmBATask):
    """
    Download raw deletions daily. At least try, files may be missing, refs
    #12460.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def run(self):
        output = shellout("""curl --fail "{url}" > {output} """,
                          url=self.config.get('vkfilmba', 'deletions'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        """
        We expect txt, but to guarantee.
        """
        return luigi.LocalTarget(path=self.path(ext="txt"))


class VKFilmBADump(VKFilmBATask):
    """
    Download pre-built initial dump,
    https://speicherwolke.uni-leipzig.de/index.php/s/KMUldvGMJRc7iLP - XXX:
        configure this out.

    Only a single fixed item, no date.
    """

    def run(self):
        output = shellout("""curl -sL --fail "https://speicherwolke.uni-leipzig.de/index.php/s/KMUldvGMJRc7iLP/download" > {output}""")
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))



class VKFilmBAUpdates(VKFilmBATask):
    """
    Iterate over download directory itself and concat items and deletions.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'dump': VKFilmBADump(),
            'data': VKFilmBADownload(date=self.date),
            'deletions': VKFilmBADownloadDeletions(date=self.date),
        }

    def run(self):
        """
        Iterate over all zipfiles in reverse, convert and concat binary marc
        into tempfile, then deduplicate.
        """

        # Load all deletions into set.
        deleted = set()

        deldir = os.path.dirname(self.input().get('deletions').path)
        for path in sorted(iterfiles(deldir), reverse=True):
            with open(path) as handle:
                for i, line in enumerate(handle, start=1):
                    line = line.strip()
                    if len(line) > 20:
                        self.logger.warn("suspicious id: %s", line)
                    deleted.add(line)

        # Load updates.
        pattern = re.compile(r'^date-[0-9]{4,4}-[0-9]{2,2}-[0-9]{2,2}.zip$')
        datadir = os.path.dirname(self.input().get('data').path)

        # Combine all binary MARC records in this file.
        _, combined = tempfile.mkstemp(prefix='siskin-')

        for path in sorted(iterfiles(datadir), reverse=True):
            filename = os.path.basename(path)

            if not pattern.match(filename):
                self.logger.warn("ignoring invalid filename: %s", path)
                continue
            if os.stat(path).st_size < 22:
                self.logger.warn("ignoring possibly empty zip file: %s", path)
                continue

            with zipfile.ZipFile(path) as zf:
                for name in zf.namelist():
                    with zf.open(name) as handle:
                        with tempfile.NamedTemporaryFile(delete=False) as dst:
                            shutil.copyfileobj(handle, dst)
                        shellout("yaz-marcdump -i marcxml -o marc {input} >> {output}",
                                 input=dst.name, output=combined, ignoremap={5: 'expected error from yaz'})
                        os.remove(dst.name)

        # Finally, concatenate initial dump.
        shellout("cat {input} >> {output}", input=self.input().get('dump').path, output=combined)

        # Already seen identifier.
        seen = set()

        with self.output().open('wb') as output:
            writer = pymarc.MARCWriter(output)

            # Iterate over MARC records (which are newest to oldest, keep track of seen identifiers).
            with open(combined) as handle:
                reader = pymarc.MARCReader(handle, force_utf8=True, to_unicode=True)
                for record in reader:
                    field = record["001"]
                    if not field:
                        self.logger.debug("missing identifier")
                        continue

                    id = field.value()

                    if id in seen:
                        self.logger.debug("skipping duplicate: %s", id)
                        continue
                    if id in deleted:
                        self.logger.debug("skipping deleted: %s", id)
                        continue

                    self.logger.debug("adding %s", id)
                    writer.write(record)
                    seen.add(id)

        self.logger.debug("found %s unique records (deletion list contained %s ids)", len(seen), len(deleted))
        os.remove(combined)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="mrc"))


class VKFilmBAMARC(VKFilmBATask):
    """
    Run conversion script.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return VKFilmBAUpdates(date=self.date)

    def run(self):
        output = shellout("python {script} {input} {output}",
                          script=self.assets("148/148_marcbinary.py"),
                          input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='fincmarc.mrc'))



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
[jstor]

ftp-host = host.name
ftp-username = username
ftp-password = password
ftp-path = /
ftp-pattern = *
"""

import collections
import datetime
import itertools
import os
import pipes
import tempfile

import luigi
import ujson as json
from gluish.format import TSV, Gzip
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout

from siskin.benchmark import timed
from siskin.common import Executable, FTPMirror
from siskin.sources.amsl import AMSLFilterConfig, AMSLService
from siskin.task import DefaultTask
from siskin.utils import SetEncoder, nwise, load_set_from_file

class JstorTask(DefaultTask):
    """ Jstor base. """
    TAG = '55'

    def closest(self):
        return weekly(self.date)

class JstorPaths(JstorTask):
    """
    Sync.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    max_retries = luigi.IntParameter(default=10, significant=False)
    timeout = luigi.IntParameter(
        default=20, significant=False, description='timeout in seconds')

    def requires(self):
        return FTPMirror(host=self.config.get('jstor', 'ftp-host'),
                         username=self.config.get('jstor', 'ftp-username'),
                         password=self.config.get('jstor', 'ftp-password'),
                         pattern=self.config.get('jstor', 'ftp-pattern'),
                         max_retries=self.max_retries,
                         timeout=self.timeout)

    def run(self):
        self.input().move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class JstorMembers(JstorTask):
    """
    Extract a full list of archive members.
    TODO: This should only be done once per file.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return JstorPaths(date=self.date)

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                if not row.path.endswith('.zip'):
                    self.logger.debug('skipping: %s', row.path)
                    continue
                shellout(""" unzip -l {input} | LC_ALL=C grep "xml$" | LC_ALL=C awk '{{print "{input}\t"$4}}' | LC_ALL=C sort >> {output} """,
                         preserve_whitespace=True, input=row.path, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class JstorLatestMembers(JstorTask):
    """
    Find the latest archive members for all article ids.

    The path names are like XXX_YYYY-MM-DD-HH-MM-SS/jjjj/yyyy-mm-dd/id/id.xml.

    This way, it is possible to sort the shipments by date, run `tac` and
    only keep the latest entries.

    2018-03-10: There is a slight change in format, refs #12669.

    Previously, update would ship in folders starting with MMS:

        MMS_2015-03-23-18-04-17/...

    with each folder containing a couple of ZIP files containing a couple of XML files.

        MMS_2015-03-23-18-04-17/ocealing_20170904T133503_1.zip

    These XML files are listed in the second column of this tasks' output.

        $ unzip -l MMS_2017-09-04-13-31-15/ocealing_20170904T133503_1.zip
        Archive:  MMS_2017-09-04-13-31-15/ocealing_20170904T133503_1.zip
        Length      Date    Time    Name
        ---------  ---------- -----   ----
            3053  2017-09-04 13:35   ocealing/2002-06-01/3623336/3623336.xml
            3392  2017-09-04 13:35   ocealing/2002-06-01/3623338/3623338.xml
            4960  2017-09-04 13:35   ocealing/2002-06-01/3623326/3623326.xml
            3002  2017-09-04 13:35   ocealing/2002-06-01/3623335/3623335.xml
            4815  2017-09-04 13:35   ocealing/2002-06-01/3623327/3623327.xml
            3980  2017-09-04 13:35   ocealing/2002-06-01/3623330/3623330.xml
            2463  2017-09-04 13:35   ocealing/2002-06-01/3623339/3623339.xml
            2814  2017-09-04 13:35   ocealing/2002-06-01/3623325/3623325.xml
            2425  2017-09-04 13:35   ocealing/2002-06-01/3623340/3623340.xml
            3170  2017-09-04 13:35   ocealing/2002-06-01/3623337/3623337.xml
            3973  2017-09-04 13:35   ocealing/2002-06-01/3623329/3623329.xml
            3874  2017-09-04 13:35   ocealing/2002-06-01/3623332/3623332.xml
            3051  2017-09-04 13:35   ocealing/2002-06-01/3623334/3623334.xml
            3536  2017-09-04 13:35   ocealing/2002-06-01/3623333/3623333.xml
            2423  2017-09-04 13:35   ocealing/2002-06-01/3623324/3623324.xml
            3600  2017-09-04 13:35   ocealing/2002-06-01/3623331/3623331.xml
            4254  2017-09-04 13:35   ocealing/2002-06-01/3623328/3623328.xml
        ---------                     -------
            58785                     17 files

    As of 2018-03-13 there are 30 new zip files in the root folder, 29 of which follow the naming scheme:

        jstor-journals-2018-03-10T06-52-44Z-part-[0-9]{3,3}.zip

    And one:

        jstor-journals-csp-2018-03-10T17-46-45Z-part-001.zip

    This tasks picked up all the files, which is why the number of records increased:

        $ taskcat JstorLatestMembers --date 2018-03-05 | wc -l
        10290925

        $ taskcat JstorLatestMembers --date 2018-03-13 | wc -l
        22136409

    One simple way to distinguish these sets:

        $ taskcat JstorLatestMembers --date 2018-03-05 | grep -c metadata
        0

        $ taskcat JstorLatestMembers --date 2018-03-13 | grep -c metadata
        11845484

        $ echo 22136409-10290925 | bc -l
        11845484

    We assume, this is a complete set of the content, so we can choose either the one or the other.

    Assumed update handling. As of 2018-07-01, there are five updates files:

        jstor-journals-updates-csp-2018-06-28T20-35-03Z-part-001.zip
        jstor-journals-updates-csp-2018-06-29T13-06-49Z-part-001.zip
        jstor-journals-updates-archives-2018-06-28T19-38-43Z-part-001.zip
        jstor-journals-updates-archives-2018-06-28T19-38-43Z-part-002.zip
        jstor-journals-updates-archives-2018-06-29T13-07-02Z-part-001.zip
        jstor-journals-updates-archives-2018-06-30T13-00-03Z-part-001.zip

    These files range in size from a few kB to almost a GB.

    XXX: Adjust snapshotting accordingly.

    Issue, refs #12669.
    """

    date = ClosestDateParameter(default=datetime.date.today())
    version = luigi.IntParameter(default=2, description="#12669")

    def requires(self):
        return JstorMembers(date=self.date)

    @timed
    def run(self):
        """
        Expect input to be sorted by shipment date, so tac will actually be a perfect rewind.
        """
        if self.version not in (1, 2):
            raise ValueError("supported versions: 1, 2 (refs #12669)")

        if self.version == 1:
            output = shellout("tac {input} | grep -v metadata | LC_ALL=C sort -S 35% -u -k2,2 | LC_ALL=C sort -S 35% -k1,1 > {output}",
                              input=self.input().path)
        if self.version == 2:
            output = shellout("tac {input} | grep metadata | LC_ALL=C sort -S 35% -u -k2,2 | LC_ALL=C sort -S 35% -k1,1 > {output}",
                              input=self.input().path)

        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class JstorXML(JstorTask):
    """
    Create a snapshot of the latest data. Using unzippa[1] to speed up
    extraction.

    [1] https://github.com/miku/unzippa
    """
    date = ClosestDateParameter(default=datetime.date.today())
    version = luigi.IntParameter(default=2, description="#12669")

    def requires(self):
        return JstorLatestMembers(date=self.date, version=self.version)

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with self.input().open() as handle:
            groups = itertools.groupby(handle.iter_tsv(
                cols=('archive', 'member')), lambda row: row.archive)

            for archive, items in groups:
                # Write members to extract to temporary file.
                _, memberfile = tempfile.mkstemp(prefix='siskin-')
                with open(memberfile, 'w') as output:
                    for item in items:
                        output.write("%s\n" % item.member)

                self.logger.debug("for archive %s extract via: %s", archive, memberfile)

                # The unzippa will not exhaust ARG_MAX.
                shellout("""unzippa -v -m {memberfile} {archive} |
                            sed -e 's@<?xml version="1.0" encoding="UTF-8"?>@@g' | pigz -c >> {output}""",
                         archive=archive.decode(encoding='utf-8'), memberfile=memberfile, output=stopover)

                try:
                    os.remove(output.name)
                except OSError as err:
                    self.logger.warn(err)

            luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml.gz'), format=TSV)

class JstorIntermediateSchemaGenericCollection(JstorTask):
    """
    Convert to intermediate format via span. Use generic JSTOR collection name.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'span': Executable(name='span-import', message='http://git.io/vI8NV'),
            'file': JstorXML(date=self.date)
        }

    @timed
    def run(self):
        output = shellout("span-import -i jstor <(unpigz -c {input}) | pigz -c > {output}",
                          input=self.input().get('file').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)

class JstorCollectionMapping(JstorTask):
    """
    Create a mapping from ISSN to collection name.

    Experimental, refs #11467. See: http://www.jstor.org/kbart/collections/all-archive-titles

        {
        "1957-7745": [
            "Arts & Sciences XIV Collection"
        ],
        "2159-4538": [
            "Arts & Sciences XII Collection"
        ],
        "0272-5045": [
            "Law Discipline Package",
            "American Society of International Law Package",
            "Arts & Sciences VI Collection"
        ],
        ...

    The KBART fields are:

     0 publication_title
     1 print_identifier
     2 online_identifier
     3 date_first_issue_online
     4 num_first_vol_online
     5 num_first_issue_online
     6 date_last_issue_online
     7 num_last_vol_online
     8 num_last_issue_online
     9 title_url
    10 first_author
    11 title_id
    12 embargo_info
    13 coverage_depth
    14 notes
    15 publisher_name
    16 publication_type
    17 date_monograph_published_print
    18 date_monograph_published_online
    19 monograph_volume
    20 monograph_edition
    21 first_editor
    22 parent_publication_title_id
    23 preceding_publication_title_id
    24 access_type
    25 full_coverage
    26 collection
    27 discipline
    28 catalog_identifier_oclc
    29 catalog_identifier_lccn
    30 archive_release_date
    31 publication_date

    """
    date = ClosestDateParameter(default=datetime.date.today())

    @timed
    def run(self):
        names = collections.defaultdict(set)
        url = "http://www.jstor.org/kbart/collections/all-archive-titles"
        output = shellout("""curl -sL "{url}" > {output} """, url=url)

        with luigi.LocalTarget(output, format=TSV).open() as handle:
            for row in handle.iter_tsv():
                if len(row) < 27:
                    self.logger.warn("short KBART row, skipping: %s", row)
                    continue

                issns, parts = row[1:3], [p.strip()
                                           for p in row[26].split(";")]
                for issn in [v.strip() for v in issns]:
                    if not issn:
                        continue
                    for name in parts:
                        if not name:
                            continue
                        names[issn].add(name)

        with self.output().open('w') as output:
            import json  # ujson does not support cls keyword
            json.dump(names, output, cls=SetEncoder)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='json'))

class JstorIntermediateSchema(JstorTask):
    """
    Turn single collection name "JStor" (https://git.io/vdHYh) into finer
    grained names via title lists (https://is.gd/W37Uwg), refs #11467.

    @2017-01-23 JSTOR ISSN to Collection mapping: https://git.io/vNwSJ
    @2017-01-23 AMSL:
        JSTOR Arts & Sciences I Archive
        JSTOR Arts & Sciences II Archive
        JSTOR Arts & Sciences III Archive
        JSTOR Arts & Sciences IV Archive
        JSTOR Arts & Sciences IX Archive
        JSTOR Arts & Sciences V Archive
        JSTOR Arts & Sciences VI Archive
        JSTOR Arts & Sciences VII Archive
        JSTOR Arts & Sciences VIII Archive
        JSTOR Arts & Sciences X Archive
        JSTOR Arts & Sciences XI Archive
        JSTOR Arts & Sciences XII Archive
        JSTOR Arts & Sciences XIII Archive
        JSTOR Arts & Sciences XIV Archive
        JSTOR Arts & Sciences XV Archive
        JSTOR Business I Archive
        JSTOR Business II Archive
        JSTOR Film and Performing Arts
        JSTOR Language & Literature Archive
        JSTOR Life Sciences Archive
        JSTOR Music Archive

    XXX: Why is JSTOR (about 500 records) still attached for DE-15, refs #12066.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'file': JstorIntermediateSchemaGenericCollection(date=self.date),
            'mapping': JstorCollectionMapping(date=self.date),
            'amsl': AMSLService(date=self.date),
        }

    @timed
    def run(self):
        """
        Only use collection names, which we find in AMSL as well. XXX: FIX deviations.
        """
        output = shellout("""unpigz -c {input} | jq -rc '.[] | select(.sourceID == "55") | .megaCollection' > {output} """,
                          input=self.input().get("amsl").path)
        allowed_collection_names = load_set_from_file(output)
        self.logger.debug("allowing via AMSL: %s", allowed_collection_names)

        # XXX: Hack to map original JSTOR collection names to names in AMSL. We
        # will need an authoritative source of such names.
        jstor_amsl_collection_name_mapping = {
            'Arts & Sciences I Collection': 'JSTOR Arts & Sciences I Archive',
            'Arts & Sciences II Collection': 'JSTOR Arts & Sciences II Archive',
            'Arts & Sciences III Collection': 'JSTOR Arts & Sciences III Archive',
            'Arts & Sciences IV Collection': 'JSTOR Arts & Sciences IV Archive',
            'Arts & Sciences IX Collection': 'JSTOR Arts & Sciences IX Archive',
            'Arts & Sciences V Collection': 'JSTOR Arts & Sciences V Archive',
            'Arts & Sciences VI Collection': 'JSTOR Arts & Sciences VI Archive',
            'Arts & Sciences VII Collection': 'JSTOR Arts & Sciences VII Archive',
            'Arts & Sciences VIII Collection': 'JSTOR Arts & Sciences VIII Archive',
            'Arts & Sciences X Collection': 'JSTOR Arts & Sciences X Archive',
            'Arts & Sciences XI Collection': 'JSTOR Arts & Sciences XI Archive',
            'Arts & Sciences XII Collection': 'JSTOR Arts & Sciences XII Archive',
            'Arts & Sciences XIII Collection': 'JSTOR Arts & Sciences XIII Archive',
            'Arts & Sciences XIV Collection': 'JSTOR Arts & Sciences XIV Archive',
            'Arts & Sciences XV Collection': 'JSTOR Arts & Sciences XV Archive',
            'Business I Collection': 'JSTOR Business I Archive',
            'Business II Collection': 'JSTOR Business II Archive',
            'Film & Performing Arts Discipline Package': 'JSTOR Film and Performing Arts',
            'Language & Literature Discipline Package': 'JSTOR Language & Literature Archive',
            'Life Sciences Collection': 'JSTOR Life Sciences Archive',
            'Music Collection': 'JSTOR Music Archive',
        }

        with self.input().get('mapping').open() as mapfile:
            mapping = json.load(mapfile)

        counter = collections.Counter()

        with self.input().get('file').open() as handle:
            with self.output().open('w') as output:
                for line in handle:
                    doc = json.loads(line)
                    issns, names = set(), set()

                    for issn in doc.get('rft.issn', []):
                        issns.add(issn)
                    for issn in doc.get('rft.eissn', []):
                        issns.add(issn)

                    for issn in issns:
                        for name in mapping.get(issn, []):
                            names.add(name)

                    if len(names) > 0:
                        # Translate JSTOR names to AMSL. TODO(miku): Get rid of special cases here.
                        amsl_names = [jstor_amsl_collection_name_mapping.get(name) for name in names if name in jstor_amsl_collection_name_mapping]
                        # Check validity against AMSL names.
                        clean_names = [name for name in amsl_names if name in allowed_collection_names]

                        # Use names that appear in AMSL.
                        doc['finc.mega_collection'] = clean_names

                        if len(doc['finc.mega_collection']) == 0:
                            self.logger.warn("no collection name given to %s: %s", doc["finc.id"], names)
                            counter["err.collection.not.in.amsl"] += 1
                    else:
                        self.logger.warn("JSTOR record without issn or issn mapping: %s", doc.get("finc.id"))
                        counter["err.name"] += 1

                    json.dump(doc, output)
                    output.write("\n")

        self.logger.debug(counter)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)

class JstorExport(JstorTask):
    """
    Tag with ISILs, then export to various formats.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='solr5vu3', description='export format')

    def requires(self):
        return {
            'file': JstorIntermediateSchema(date=self.date),
            'config': AMSLFilterConfig(date=self.date),
        }

    def run(self):
        output = shellout("span-tag -c {config} <(unpigz -c {input}) | pigz -c > {output}",
                          config=self.input().get('config').path, input=self.input().get('file').path)
        output = shellout(
            "span-export -o {format} <(unpigz -c {input}) | pigz -c > {output}", format=self.format, input=output)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        extensions = {
            'solr5vu3': 'ldj.gz',
            'formeta': 'form.gz',
        }
        return luigi.LocalTarget(path=self.path(ext=extensions.get(self.format, 'gz')))

class JstorISSNList(JstorTask):
    """
    A list of JSTOR ISSNs.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return JstorIntermediateSchema(date=self.date)

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        shellout(
            """jq -r '.["rft.issn"][]?' <(unpigz -c {input}) 2> /dev/null >> {output} """, input=self.input().path, output=stopover)
        shellout(
            """jq -r '.["rft.eissn"][]?' <(unpigz -c {input}) 2> /dev/null >> {output} """, input=self.input().path, output=stopover)
        output = shellout("""sort -u {input} > {output} """, input=stopover)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class JstorDOIList(JstorTask):
    """
    A list of JSTOR DOIs.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return JstorIntermediateSchema(date=self.date)

    @timed
    def run(self):
        output = shellout("""jq -r '.doi' <(unpigz -c {input}) | grep -v null > {output} """,
                          input=self.input().path)
        output = shellout("""sort -u {input} > {output} """, input=output)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301,C0330

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

import datetime

import luigi
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout

from siskin.sources.amsl import AMSLFilterConfig
from siskin.task import DefaultTask


"""
IJOC, refs #7138, #11005.
"""


class IJOCTask(DefaultTask):
    """ Base task for International Journal of Communication """
    TAG = '87'

    def closest(self):
        return monthly(date=self.date)


class IJOCHarvest(IJOCTask):
    """
    Harvest.
    """
    endpoint = luigi.Parameter(
        default='http://ijoc.org/index.php/ijoc/oai', significant=False)
    date = ClosestDateParameter(default=datetime.date.today())

    def run(self):
        shellout("""METHA_DIR={dir} metha-sync "{endpoint}" """,
                 dir=self.config.get('core', 'metha-dir'),
                 endpoint=self.endpoint)
        output = shellout("""METHA_DIR={dir} metha-cat -root Records "{endpoint}" > {output}""",
                          dir=self.config.get('core', 'metha-dir'),
                          endpoint=self.endpoint)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml'))


class IJOCIntermediateSchema(IJOCTask):
    """
    Convert to intermediate schema via metafacture. Custom morphs and flux are kept in assets/87.
    Maps are kept in assets/maps
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return IJOCHarvest(date=self.date)

    def run(self):
        mapdir = 'file:///%s' % self.assets("maps/")
        output = shellout("""flux.sh {flux} in={input} MAP_DIR={mapdir} | pigz -c > {output}""",
                          flux=self.assets("87/87.flux"), mapdir=mapdir, input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))


class IJOCFincSolr(IJOCTask):
    """
    Export to finc solr schema by using span-export.
    Tag with ISIL for FID and change record type.
    """
    format = luigi.Parameter(default='solr5vu3', description='export format')
    isil = luigi.Parameter(default='DE-15-FID', description='isil FID')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'config': AMSLFilterConfig(date=self.date),
            'file': IJOCIntermediateSchema(date=self.date)
        }

    def run(self):
        output = shellout("""span-tag -c {config} <(unpigz -c {input}) | span-export -o {format} -with-fullrecord > {output}""",
                          config=self.input().get('config').path, input=self.input().get('file').path,
                          format=self.format)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='fincsolr.ndj'))



from siskin.openurl import openurl_parameters_from_intermediateschema

def test_openurl_from_intermediateschema():
    cases = (
        ('empty doc', {}, {}),
        (
            'title only',
            {'rft.atitle': 'empty doc'},
            {
                'ctx_enc': 'info:ofi/enc:UTF-8',
                'ctx_ver': 'Z39.88-2004',
                'rfr_id': 'info:sid/www.ub.uni-leipzig.de:generator',
                'rft.atitle': 'empty doc',
                'url_ver': 'Z39.88-2004',
            },
        ),
        (
            'title and date',
            {
                'rft.atitle': 'title and date',
                'rft.date': '2018-10-10',
            },
            {
                'ctx_enc': 'info:ofi/enc:UTF-8',
                'ctx_ver': 'Z39.88-2004',
                'rfr_id': 'info:sid/www.ub.uni-leipzig.de:generator',
                'rft.date': '2018-10-10',
                'rft.atitle': 'title and date',
                'url_ver': 'Z39.88-2004',
            },
        ),
        (
            'title and date, language',
            {
                'languages': ['eng', 'fra'],
                'rft.atitle': 'title and date, language',
                'rft.date': '2018-10-10',
            },
            {
                'ctx_enc': 'info:ofi/enc:UTF-8',
                'ctx_ver': 'Z39.88-2004',
                'rfr_id': 'info:sid/www.ub.uni-leipzig.de:generator',
                'rft.date': '2018-10-10',
                'rft.language': 'eng',
                'rft.atitle': 'title and date, language',
                'url_ver': 'Z39.88-2004',
            },
        ),
        (
            'title and date, language, book',
            {
                'languages': ['eng', 'fra'],
                'rft.atitle': 'Hello',
                'rft.date': '2018-10-10',
                'rft.genre': 'book',
            },
            {
                'ctx_enc': 'info:ofi/enc:UTF-8',
                'ctx_ver': 'Z39.88-2004',
                'rfr_id': 'info:sid/www.ub.uni-leipzig.de:generator',
                'rft.atitle': 'Hello',
                'rft.btitle': 'Hello',
                'rft.date': '2018-10-10',
                'rft.genre': 'book',
                'rft.language': 'eng',
                'rft_val_fmt': 'info:ofi/fmt:kev:mtx:book',
                'url_ver': 'Z39.88-2004',
            },
        ),
        (
            'crossref-1',
            {
              "finc.format": "ElectronicArticle",
              "finc.mega_collection": [
                "Springer Science + Business Media (CrossRef)"
              ],
              "finc.id": "ai-49-aHR0cDovL2R4LmRvaS5vcmcvMTAuMTAxNi9qLm51cnguMjAwNi4wNS4wMjU",
              "finc.source_id": "49",
              "ris.type": "EJOUR",
              "rft.atitle": "An Analysis of Correlations Among 4 Outcome Scales Employed in Clinical Trials of Patients With Major Depressive Disorder",
              "rft.epage": "412",
              "rft.genre": "article",
              "rft.issn": [
                "1545-5343"
              ],
              "rft.issue": "3",
              "rft.jtitle": "NeuroRX",
              "rft.tpages": "2",
              "rft.pages": "411-412",
              "rft.pub": [
                "Springer Science + Business Media"
              ],
              "rft.date": "2006-07-01",
              "x.date": "2006-07-01T00:00:00Z",
              "rft.spage": "411",
              "rft.volume": "3",
              "authors": [
                {
                  "rft.aulast": "JIANG",
                  "rft.aufirst": "Q"
                },
                {
                  "rft.aulast": "AHMED",
                  "rft.aufirst": "S"
                },
                {
                  "rft.aulast": "PEDERSEN",
                  "rft.aufirst": "R"
                },
                {
                  "rft.aulast": "MUSGNUNG",
                  "rft.aufirst": "J"
                },
                {
                  "rft.aulast": "ENTSUAH",
                  "rft.aufirst": "R"
                }
              ],
              "doi": "10.1016/j.nurx.2006.05.025",
              "languages": [
                "eng"
              ],
              "url": [
                "http://dx.doi.org/10.1016/j.nurx.2006.05.025"
              ],
              "version": "0.9",
              "x.subjects": [
                "Pharmacology (medical)"
              ],
              "x.type": "journal-article"
            },
            {
                'ctx_enc': 'info:ofi/enc:UTF-8',
                'ctx_ver': 'Z39.88-2004',
                'rfr_id': 'info:sid/www.ub.uni-leipzig.de:generator',
                'rft.atitle': 'An Analysis of Correlations Among 4 Outcome Scales Employed in Clinical Trials of Patients With Major Depressive Disorder',
                'rft.aufirst': 'Q',
                'rft.aulast': 'JIANG',
                'rft.date': '2006-07-01',
                'rft.epage': '412',
                'rft.genre': 'article',
                'rft.issn': '1545-5343',
                'rft.issue': '3',
                'rft.jtitle': 'NeuroRX',
                'rft.language': 'eng',
                'rft.pages': '411-412',
                'rft.spage': '411',
                'rft.volume': '3',
                'rft_id': 'info:doi/10.1016/j.nurx.2006.05.025',
                'url_ver': 'Z39.88-2004',
            },
        ),
    )

    for _, doc, want in cases:
        result = openurl_parameters_from_intermediateschema(doc)
        assert result == want

